package org.bdware.consistency.plugin.ra;

import com.google.gson.JsonObject;
import com.google.gson.JsonParser;
import com.google.gson.JsonPrimitive;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.bdware.consistency.plugin.common.AbstractContextContractExecutor;
import org.bdware.sc.ComponedContractResult;
import org.bdware.sc.ContractResult;
import org.bdware.sc.bean.ContractExecType;
import org.bdware.sc.bean.ContractRequest;
import org.bdware.sc.conn.OnHashCallback;
import org.bdware.sc.conn.ResultCallback;
import org.bdware.sc.units.MultiContractMeta;
import org.bdware.sc.units.RecoverFlag;
import org.bdware.sc.units.RequestCache;
import org.bdware.sc.units.ResultCache;
import org.bdware.sc.util.JsonUtil;
import org.bdware.sdk.consistency.api.NotifiableResultMerger;
import org.bdware.server.trustedmodel.MultiReqSeq;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;

public class RequestAllExecutor extends AbstractContextContractExecutor {
    private static final Logger LOGGER = LogManager.getLogger(RequestAllExecutor.class);
    final Object lock = new Object();
    int resultCount;
    AtomicInteger request_index = new AtomicInteger(0);
    ContractExecType type;
    // key为requestID，value为其seq
    Map<String, MultiReqSeq> seqMap = new ConcurrentHashMap<>();
    Map<String, ResultCache> resultCache = new ConcurrentHashMap<>();
    //  MultiPointContractInfo info;
    String contractID;

    public RequestAllExecutor(
            ContractExecType t, int c, String con_id) {
        type = t;
        resultCount = c;
        contractID = con_id;
    }

    public void setSeq(int seq) {
        request_index = new AtomicInteger(seq);
    }

    public ResultCallback createResultCallback(
            final String requestID,
            final ResultCallback originalCb,
            final int count,
            final int request_seq,
            final String contractID) {
        ComponedContractResult componedContractResult = new ComponedContractResult(count);
        // TODO 加对应的超时？
        return networkManager.createResultCallback(
                requestID, new ResultMerger(originalCb, count, request_seq, contractID), count);
    }

    public void sendRequest(String id, ContractRequest req, ResultCallback collector) {
        Map<String, Object> reqStr = new HashMap<>();
        reqStr.put("uniReqID", id);
        reqStr.put("data", req);
        reqStr.put("action", "executeContractLocally");
        String sendStr = JsonUtil.toJson(reqStr);

        // master负责缓存请求
        if (!masterServerTCPAction.getReqCache().containsKey(contractID)) {
            masterServerTCPAction.getReqCache().put(contractID, new RequestCache());
        }
        // TODO 多调多统一个seq的有多个请求，这个需要改
        masterServerTCPAction.getReqCache().get(contractID).put(req.seq, sendStr);

        LOGGER.debug(JsonUtil.toJson(req));

        String[] nodes =
                cmActions.getManager().multiContractRecorder.getMultiContractMeta(contractID).getMembers();
        LOGGER.info("cluster size=" + nodes.length + " contract " + req.getContractID());
        LOGGER.debug("contract " + req.getContractID() + " cluster: " + JsonUtil.toJson(nodes));
        for (String node : nodes) {
            LOGGER.debug("get cmNode " + node.substring(0, 5));
            if (!networkManager.hasAgentConnection(node)) {
                LOGGER.warn("cmNode " + node.substring(0, 5) + " is null");
                collector.onResult(
                        "{\"status\":\"Error\",\"result\":\"node offline\","
                                + "\"nodeID\":\""
                                + node
                                + "\","
                                + "\"action\":\"onExecuteContractTrustfully\"}");
            } else if (masterServerRecoverMechAction.getRecoverStatusMap().get(node).get(contractID)
                    != RecoverFlag.Fine) {
                collector.onResult(
                        "{\"status\":\"Error\",\"result\":\"node recovering\","
                                + "\"nodeID\":\""
                                + node
                                + "\","
                                + "\"action\":\"onExecuteContractTrustfully\"}");
                networkManager.sendToAgent(node, sendStr);
            } else {
                LOGGER.info("send request to cmNode " + node.substring(0, 5));
                networkManager.sendToAgent(node, sendStr);
            }
        }
    }

    public boolean checkCurNodeNumValid() {
        String[] nodes =
                cmActions.getManager().multiContractRecorder.getMultiContractMeta(contractID).getMembers();
        int validNode = 0;
        Map<String, String> mapResult = new HashMap<>();
        for (String node : nodes) {
            mapResult.put(node.substring(0, 5), String.format("%s %s", networkManager.hasAgentConnection(node) + "",
                    masterServerRecoverMechAction.getRecoverStatusMap().get(node).get(contractID)));
            if (networkManager.hasAgentConnection(node)
                    && masterServerRecoverMechAction.getRecoverStatusMap().get(node).get(contractID)
                    == RecoverFlag.Fine) {
                validNode++;
            }
        }
        LOGGER.info(JsonUtil.toPrettyJson(mapResult));
        int c = resultCount;
        if (ContractExecType.RequestAllResponseAll.equals(type)) {
            c = (int) Math.ceil((double) c / 2);
        }
        LOGGER.debug("c=" + c + " validNode=" + validNode);
        return validNode >= c;
    }

    @Override
    public void execute(String requestID, ContractRequest req, ResultCallback rc, OnHashCallback hcb) {
        LOGGER.debug(JsonUtil.toJson(req));
        MultiContractMeta meta = cmActions.getManager().multiContractRecorder.getMultiContractMeta(req.getContractID());
        if (meta == null || !meta.isMaster()) {
            cmActions.getManager().executeContractOnOtherNodes(req, rc);
            return;
        }
        req.setContractID(cmActions.getManager().getContractIDByName(req.getContractID()));

        // 三个相同requestID进来的时候，会有冲突。
        // 仅在此处有冲突么？
        // 这里是从MasterServer->MasterClient，请求的是"executeContractLocally"。

        // 如果是多点合约的请求，A1、A2、A3的序号应该一致，不能分配一个新的seq，根据requestID判断是否不需要重新分配一个序号
        //TODO seqMap memory leak
        //TODO
        //TODO
        if (null != requestID && requestID.endsWith("_mul")) {
            synchronized (lock) {
                if (seqMap.containsKey(requestID)) {
                    req.seq = seqMap.get(requestID).seq;
                } else {
                    req.seq = request_index.getAndIncrement();
                    seqMap.put(requestID, new MultiReqSeq(req.seq));
                }
            }
        } else {
            req.seq = request_index.getAndIncrement();
        }
        req.needSeq = true;
        String id =
                System.currentTimeMillis() + "_" + (int) (Math.random() * 1000000) + "_" + req.seq;
        LOGGER.info("execute receive requestID=" + requestID + "  msgID=" + id);

        if (checkCurNodeNumValid()) {
            LOGGER.debug("checkCurNodeNumValid=true");
            ResultCallback collector =
                    createResultCallback(id, rc, resultCount, req.seq, req.getContractID());
            masterServerTCPAction.getSync().sleep(id, collector);
            LOGGER.info("requestID=" + requestID + "  master broadcasts request " + req.seq);
            sendRequest(id, req, collector);
        } else {
            LOGGER.debug("invalidNodeNumOnResult");
            request_index.getAndDecrement();
            ContractResult finalResult =
                    new ContractResult(
                            ContractResult.Status.Error,
                            new JsonPrimitive("node number unavailable, request refused."));
            rc.onResult(JsonUtil.toJson(finalResult));
        }

        // }

        /*        // 三个相同requestID进来的时候，会有冲突。
        // 仅在此处有冲突么？
        // 这里是从MasterServer->MasterClient，请求的是"executeContractLocally"。
        req.seq = request_index.getAndIncrement();
        req.needSeq = true;
        ResultCallback collector = createResultCallback(id, rc, resultCount, req.getContractID());
        MasterServerTCPAction.sync.sleep(id, collector);
        sendRequest(id, req, collector);*/
    }

    // 清理缓存的多点合约请求序号
    public void clearCache() {
        final long time = System.currentTimeMillis() - 30000L;
        seqMap.entrySet()
                .removeIf(
                        entry -> {
                            MultiReqSeq cache = entry.getValue();
                            if (null == cache) {
                                return true;
                            }
                            return cache.startTime < time;
                        });
    }

    public static class ResultMerger extends ResultCallback implements NotifiableResultMerger {
        ComponedContractResult componedContractResult;
        AtomicInteger order;
        String contractID;
        int count;
        int request_seq;
        ResultCallback originalCallback;
        Set<String> nodeIDs = new HashSet<>(); // 已收到返回结果的节点

        ResultMerger(
                final ResultCallback originalCb,
                final int count,
                final int request_seq,
                final String contractID) {
            originalCallback = originalCb;
            this.count = count;
            this.request_seq = request_seq;
            this.contractID = contractID;
            componedContractResult = new ComponedContractResult(count);
            order = new AtomicInteger(0);
        }

        public String getContractID() {
            return contractID;
        }

        public String getInfo() {
            return "contractID="
                    + contractID
                    + "  收到第 "
                    + order
                    + " 个节点回复 : "
                    + "  order="
                    + order
                    + "   count="
                    + count
                    + "  ";
        }

        @Override
        public void onResult(String str) {
            // TODO 必须在这里聚合。
            // str的data是个ContractResult
            // 在这儿也是返回个ContractResult
            try {
                LOGGER.debug("a result of contract" + contractID + ": " + str);
                JsonObject obj = JsonParser.parseString(str).getAsJsonObject();
                if (obj.has("nodeID")) {
                    String id = obj.get("nodeID").getAsString();
                    if (nodeIDs.contains(id)) {
                        LOGGER.debug(
                                "ignored result because the result of node "
                                        + id.substring(0, 5)
                                        + " has been received");
                        return;
                    }
                    nodeIDs.add(id);
                }

                LOGGER.debug(
                        String.format(
                                "contractID=%s  received=%s  order=%d  count=%d",
                                contractID, str, order.get(), count));
                componedContractResult.add(obj);
                // 收集到所有结果
                if (order.incrementAndGet() == count) {
                    ContractResult finalResult = componedContractResult.figureFinalResult();
                    finalResult.needSeq = true;
                    finalResult.seq = request_seq;

                    //                    if (null == finalResult) {
                    //                        finalResult =
                    //                                new ContractResult(
                    //                                        ContractResult.Status.Exception,
                    //                                        new JsonPrimitive(
                    //                                                "no nore than half of the
                    // consistent result"));
                    //                        originalCallback.onResult(new
                    // Gson().toJson(finalResult));
                    //                    } else {
                    originalCallback.onResult(JsonUtil.toJson(finalResult));
                    //                    }
                    LOGGER.debug(
                            String.format(
                                    "%d results are the same: %s",
                                    finalResult.size, finalResult.result));

                    // 集群中事务序号+1
                    cmActions.getManager().multiContractRecorder
                            .getMultiContractMeta(contractID)
                            .nextSeqAtMaster();

                    // recover，其中无状态合约CP出错无需恢复
                    Set<String> nodesID = componedContractResult.getProblemNodes();
                    if (null == nodesID || nodesID.isEmpty()) {
                        return;
                    }
                    for (String nodeID : nodesID) {
                        LOGGER.warn("node fails! " + nodeID);
                        if (masterServerRecoverMechAction.getRecoverStatusMap().get(nodeID).get(contractID)
                                == RecoverFlag.Fine) {
                            masterServerRecoverMechAction.getRecoverStatusMap()
                                    .get(nodeID)
                                    .put(contractID, RecoverFlag.ToRecover);
                        }
                    }
                    for (String nodeID : nodesID) {
                        if (masterServerRecoverMechAction.getRecoverStatusMap().get(nodeID).get(contractID)
                                == RecoverFlag.ToRecover) {
                            LOGGER.warn("node in recover " + nodeID);

                            // 因为该节点结果有误，所以即时是stableMode也认为trans记录不可信
                            // 直接通过load别的节点来恢复
                            masterServerRecoverMechAction.restartContractFromCommonMode(
                                    nodeID, contractID);
                        }
                    }
                }
                //  clearCache();
            } catch (Exception e) {
                e.printStackTrace();
                LOGGER.warn("result exception!");
            }
        }
    }

    @Override
    public void onRecover(Map<String, Object> args) {
        int ceiLastExeSeq = (int) args.get("ceiLastExeSeq");
        this.setSeq(ceiLastExeSeq + 1);
    }
}
