package org.bdware.consistency.plugin.ro;

import com.google.gson.JsonObject;
import com.google.gson.JsonParser;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.bdware.consistency.plugin.common.AbstractContextContractExecutor;
import org.bdware.sc.ContractResult;
import org.bdware.sc.bean.ContractRequest;
import org.bdware.sc.conn.OnHashCallback;
import org.bdware.sc.conn.ResultCallback;
import org.bdware.sc.util.JsonUtil;

import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.atomic.AtomicInteger;

public class ResponseOnceExecutor extends AbstractContextContractExecutor {
    private static final Logger LOGGER = LogManager.getLogger(ResponseOnceExecutor.class);
    private final String contractID;
    AtomicInteger order = new AtomicInteger(0);

    public ResponseOnceExecutor(String contractID) {
        this.contractID = contractID;
    }

    @Override
    public void execute(String requestID, ContractRequest req, ResultCallback rc, OnHashCallback hcb) {
        executeInternal(requestID, rc, req, 2);
    }

    private void executeInternal(
            String requestID, ResultCallback rc, ContractRequest req, int count) {
        //        String contractID = req.getContractID();
        // TODO 标注失效节点，是否选择重新迁移？？
        ResultCallback cb =
                new ResultCallback() {
                    @Override
                    public void onResult(String str) {
                        LOGGER.debug(str);
                        JsonObject jo = JsonParser.parseString(str).getAsJsonObject();
                        jo.remove("action");
                        jo.addProperty("action", "onExecuteResult");
                        LOGGER.debug(jo.toString());
                        if (jo.has("data")) {
                            String data = jo.get("data").getAsString();
                            ContractResult cr = JsonUtil.fromJson(data, ContractResult.class);
                            if (cr.status != ContractResult.Status.Success && count > 0) {
                                executeInternal(requestID, rc, req, count - 1);
                            } else rc.onResult(jo.toString());
                        } else {
                            JsonObject jo2 = new JsonObject();
                            jo2.addProperty("action", "onExecuteResult");
                            jo.remove("action");
                            jo2.addProperty("data", jo.toString());
                            rc.onResult(jo2.toString());
                        }
                    }
                };
        masterServerTCPAction.getSync().sleepWithTimeout(requestID, cb, 5);
        if (!sendOnce(requestID, req))
            rc.onResult(
                    "{\"status\":\"Error\",\"data\":\"{\\\"status\\\":\\\"Error\\\",\\\"result\\\":\\\"all nodes offline\\\"}\",\"action\":\"onExecuteContract\"}");
    }

    private boolean sendOnce(String requestID, ContractRequest req) {
        String[] members = cmActions.getManager().multiContractRecorder.getMultiContractMeta(contractID).getMembers();
        for (int i = 0; i < members.length; i++) {
            int size = members.length;
            String nodeID = members[order.incrementAndGet() % size];
            Map<String, Object> obj = new HashMap<>();
            obj.put("action", "executeContractLocally");
            obj.put("data", req);
            obj.put("uniReqID", requestID);
            networkManager.sendToAgent(nodeID, JsonUtil.toJson(obj));
            return true;
        }
        return false;
    }
}
