package org.bdware.consistency.plugin.sharding;

import com.google.gson.JsonObject;
import com.google.gson.JsonParser;
import com.google.gson.JsonPrimitive;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.bdware.consistency.plugin.common.AbstractContextContractExecutor;
import org.bdware.sc.ComponedContractResult;
import org.bdware.sc.ContractMeta;
import org.bdware.sc.ContractResult;
import org.bdware.sc.bean.*;
import org.bdware.sc.conn.OnHashCallback;
import org.bdware.sc.conn.ResultCallback;
import org.bdware.sc.units.MultiContractMeta;
import org.bdware.sc.units.RecoverFlag;
import org.bdware.sc.units.RequestCache;
import org.bdware.sc.units.ResultCache;
import org.bdware.sc.util.JsonUtil;
import org.bdware.server.trustedmodel.MultiReqSeq;

import java.math.BigInteger;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;

// 改为MultiPointCooperationExecutor
public class MultiPointCooperationExecutor extends AbstractContextContractExecutor {
    private static final Logger LOGGER = LogManager.getLogger(MultiPointCooperationExecutor.class);
    final Object lock = new Object();
    int resultCount;
    AtomicInteger request_index = new AtomicInteger(0);
    ContractExecType type;
    // key为requestID，value为其seq
    Map<String, MultiReqSeq> seqMap = new ConcurrentHashMap<>();
    Map<String, ResultCache> resultCache = new ConcurrentHashMap<>();
    // MultiPointContractInfo info;
    MultiContractMeta multiMeta;
    String contractID;

    public MultiPointCooperationExecutor(ContractExecType t, int c, String con_id) {
        LOGGER.info("-- sharding executor---");
        type = t;
        resultCount = c;
        contractID = con_id;
        multiMeta = cmActions.getManager().multiContractRecorder.getMultiContractMeta(contractID);
    }

    public void setSeq(int seq) {
        request_index = new AtomicInteger(seq);
    }

    public ResultCallback createResultCallback(
            final String requestID,
            final ResultCallback originalCb,
            final int count,
            final int request_seq,
            final String contractID, JoinInfo joinInfo) {
        // TODO 加对应的超时？
        return networkManager.createResultCallback(
                requestID,
                new MultiPointCooperationExecutor.ResultMerger(originalCb, count, request_seq, contractID, joinInfo),
                count); // 把count改成了1，设置成获得1个响应就行
    }

    public void sendRequest(String id, ContractRequest req, String[] nodes) {
        Map<String, Object> reqStr = new HashMap<>();
        reqStr.put("uniReqID", id);
        reqStr.put("data", req);
        req.needSeq = false;
        reqStr.put("action", "executeContractLocally");
        String sendStr = JsonUtil.toJson(reqStr);
        // master负责缓存请求
        if (!masterServerTCPAction.getReqCache().containsKey(contractID)) {
            masterServerTCPAction.getReqCache().put(contractID, new RequestCache());
        }
        // TODO 多调多统一个seq的有多个请求，这个需要改
        masterServerTCPAction.getReqCache().get(contractID).put(req.seq, sendStr);
        LOGGER.debug(JsonUtil.toJson(req));
        LOGGER.info("node size = " + nodes.length);
        LOGGER.debug("nodes:" + JsonUtil.toJson(nodes));
        for (String node : nodes) {
            LOGGER.info(
                    "[sendRequests] get cmNode "
                            + node.substring(0, 5)
                            + " not null "
                            + "RequestAllExecutor  发送请求给 "
                            + node.substring(0, 5));
            networkManager.sendToAgent(node, sendStr);
        }
    }

    private String[] getAccordingToRouteInfo(RouteInfo routeInfo, ContractRequest req, String[] members) {
        try {
            int val;
            switch (routeInfo.useDefault) {
                case byRequester:
                    val =
                            new BigInteger(req.getRequester(), 16)
                                    .mod(new BigInteger("" + members.length))
                                    .intValue();
                    while (val < 0) {
                        val = val + members.length;
                    }
                    return new String[]{members[val]};
                case byArgHash:
                    val = req.getArg().hashCode();
                    val = val % members.length;
                    while (val < 0) {
                        val += members.length;
                    }
                    return new String[]{members[val]};
                case byTarget:
                    JsonObject jo = req.getArg().getAsJsonObject();
                    val =
                            new BigInteger(jo.get("target").getAsString(), 16)
                                    .mod(new BigInteger("" + members.length))
                                    .intValue();
                    while (val < 0) {
                        val = val + members.length;
                    }
                    return new String[]{members[val]};
                default:
                    return members;
            }
        } catch (Exception e) {
            return members;
        }
    }

    public boolean checkCurNodeNumValid() {
        LOGGER.info("checkCurNodeNumValid");
        String[] nodes = multiMeta.getMembers();
        // List<String> nodes = info.members;
        int validNode = 0;
        for (String node : nodes) {
            if (networkManager.hasAgentConnection(node)) {
                //&& masterServerRecoverMechAction.getRecoverStatusMap().get(node).get(contractID)
                //                    == RecoverFlag.Fine
                validNode++;
            }
        }
        int c = resultCount;
        if (type == ContractExecType.Sharding) c = (int) Math.ceil((double) c / 2);
        LOGGER.info("c=" + c + "  validNode=" + validNode);
        return validNode >= c;
    }

    @Override
    public void execute(String requestID, ContractRequest req, ResultCallback rc, OnHashCallback hcb) {
        LOGGER.info("[MultiPointCooperationExecutor] execute " + JsonUtil.toJson(req));
        // 获得action 函数名
        LOGGER.info("action is : " + req.getAction());
        req.setContractID(cmActions.getManager().getContractIDByName(req.getContractID()));
        if (requestID != null && requestID.endsWith("_mul")) {
            synchronized (lock) {
                if (seqMap.containsKey(requestID)) {
                    req.seq = seqMap.get(requestID).seq;
                } else {
                    req.seq = request_index.getAndIncrement();
                    seqMap.put(requestID, new MultiReqSeq(req.seq));
                }
            }
        } else {
            req.seq = request_index.getAndIncrement();
        }
        req.needSeq = true;
        String id =
                System.currentTimeMillis() + "_" + (int) (Math.random() * 1000000) + "_" + req.seq;
        LOGGER.info("execute receive requestID= " + requestID + "  msgID=" + id);
        if (checkCurNodeNumValid()) { // 校验成功 current node num 合法
            LOGGER.info("checkCurNodeNumValid true");
            ContractMeta meta =
                    cmActions.getManager().statusRecorder.getContractMeta(req.getContractID());
            FunctionDesp fun = meta.getExportedFunction(req.getAction());
            ResultCallback collector;
            // TODO @fanbo 下面的count 1要改，应该是根据route的规则来。
            //Count 根据join规则来。
            //nodes 根据route规则来。
            JoinInfo joinInfo = fun.joinInfo;
            RouteInfo routeInfo = fun.routeInfo;
            int count = getJoinCount(joinInfo, contractID);
            LOGGER.info("requestID=" + requestID + " join Count: " + count);

            String[] members = multiMeta.getMembers();
            String[] nodes = getAccordingToRouteInfo(routeInfo, req, members);
            if (nodes.length < count) {
                count = nodes.length;
            }
            collector =
                    createResultCallback(id, rc, count, req.seq, req.getContractID(), joinInfo); // 初始化结果收集器
            masterServerTCPAction.getSync().sleep(id, collector);
            LOGGER.info("requestID=" + requestID + "  master broadcasts request " + req.seq);
            sendRequest(id, req, nodes); // 发送请求
        } else {
            LOGGER.info("invalidNodeNumOnResult");
            request_index.getAndDecrement();
            ContractResult finalResult =
                    new ContractResult(
                            ContractResult.Status.Error,
                            new JsonPrimitive("node number unavailbale,request refused."));
            rc.onResult(JsonUtil.toJson(finalResult));
        }
    }

    private int getJoinCount(JoinInfo joinInfo, String contractID) {
        if (joinInfo == null) return resultCount;
        if (joinInfo != null) return joinInfo.joinCount;

        try {
            ContractRequest cr = new ContractRequest();
            cr.setContractID(contractID);
            cr.setAction("TODO");
            //TODO Arg需要好好设计一下。
            //TODO 又好用又简单的那种设计
            //TODO
            cr.setArg("");
            String result = cmActions.getManager().executeLocally(cr, null);
            return JsonUtil.parseString(result).getAsJsonObject().get("result").getAsInt();
        } catch (Exception e) {
            e.printStackTrace();
            return 1;
        }
    }

    // 清理缓存的多点合约请求序号
    public void clearCache() {
        final long time = System.currentTimeMillis() - 30000L;
        seqMap.entrySet()
                .removeIf(
                        entry -> {
                            MultiReqSeq cache = entry.getValue();
                            if (null == cache) {
                                return true;
                            }
                            return cache.startTime < time;
                        });
    }

    public static class ResultMerger extends ResultCallback {
        ComponedContractResult componedContractResult;
        AtomicInteger order;
        String contractID;
        int count; // 记录有多少个节点
        int request_seq;
        ResultCallback originalCallback;
        Set<String> nodeIDs = new HashSet<>(); // 已收到返回结果的节点
        JoinInfo joinInfo;

        ResultMerger(
                final ResultCallback originalCb,
                final int count,
                final int request_seq,
                final String contractID,
                final JoinInfo joinInfo) {
            originalCallback = originalCb;
            this.count = count;
            this.request_seq = request_seq;
            this.contractID = contractID;
            componedContractResult = new ComponedContractResult(count);
            order = new AtomicInteger(0);
            this.joinInfo = joinInfo;
        }

        public String getInfo() {
            return "contractID="
                    + contractID
                    + "  收到第 "
                    + order
                    + " 个节点回复 : "
                    + "  order="
                    + order
                    + "   count="
                    + count
                    + "  ";
        }

        @Override
        public void onResult(String str) {
            // TODO 必须在这里聚合。
            // str的data是个ContractResult
            // 在这儿也是返回个ContractResult
            try {
                LOGGER.info(str);
                JsonObject obj = JsonParser.parseString(str).getAsJsonObject();
                String id = obj.get("nodeID").getAsString();
                if (nodeIDs.contains(id)) {
                    LOGGER.info("已经收到节点 " + id.substring(0, 5) + " 的结果，该结果被忽略");
                    return;
                }
                nodeIDs.add(id);
                LOGGER.info(
                        "contractID="
                                + contractID
                                + "  收到第 "
                                + order
                                + " 个节点回复 : "
                                + str
                                + "  order="
                                + order
                                + "   count="
                                + count);
                componedContractResult.add(obj);
                // 收集到所有结果
                if (order.incrementAndGet() == count) {
                    ContractResult finalResult = componedContractResult.mergeFinalResult();

                    finalResult.needSeq = true;
                    finalResult.seq = request_seq;

                    //                    if (null == finalResult) {
                    //                        finalResult =
                    //                                new ContractResult(
                    //                                        ContractResult.Status.Exception,
                    //                                        new JsonPrimitive(
                    //                                                "no nore than half of the
                    // consistent result"));
                    //                        originalCallback.onResult(new
                    // Gson().toJson(finalResult));
                    //                    } else {
                    if (joinInfo != null) {
                        handleJoinInfo(finalResult, joinInfo);
                    }
                    originalCallback.onResult(JsonUtil.toJson(finalResult));
                    //                    }
                    LOGGER.info(
                            "本次执行最终结果为 " + finalResult.size + "个节点合并的,结果为 " + finalResult.result);

                    // 集群中事务序号+1
                    // MasterServerTCPAction.contractID2Members.get(contractID).nextSeq();
                    cmActions.getManager()
                            .multiContractRecorder
                            .getMultiContractMeta(contractID)
                            .nextSeqAtMaster();
                    // recover，其中无状态合约CP出错无需恢复
                    Set<String> nodesID = componedContractResult.getProblemNodes();
                    if (null == nodesID || nodesID.isEmpty()) {
                        return;
                    }
                    for (String nodeID : nodesID) {
                        LOGGER.info("结果出现问题的节点有：" + nodeID);
                        if (masterServerRecoverMechAction.getRecoverStatusMap().get(nodeID).get(contractID)
                                == RecoverFlag.Fine) {
                            masterServerRecoverMechAction.getRecoverStatusMap()
                                    .get(nodeID)
                                    .put(contractID, RecoverFlag.ToRecover);
                        }
                    }
                    for (String nodeID : nodesID) {
                        if (masterServerRecoverMechAction.getRecoverStatusMap().get(nodeID).get(contractID)
                                == RecoverFlag.ToRecover) {
                            LOGGER.info("问题节点开始恢复：" + nodeID);

                            // 因为该节点结果有误，所以即时是stableMode也认为trans记录不可信
                            // 直接通过load别的节点来恢复
                            masterServerRecoverMechAction.restartContractFromCommonMode(
                                    nodeID, contractID);
                        }
                    }
                }
                //  clearCache();
            } catch (Exception e) {
                e.printStackTrace();
                LOGGER.info("本次执行最终结果为有异常");
            }
        }

        private void handleJoinInfo(ContractResult finalResult, JoinInfo joinInfo) {
            JsonObject jo = finalResult.result.getAsJsonObject();
            if (joinInfo != null && joinInfo.joinRule != null) {
                //TODO 不应该是double 类型
                switch (joinInfo.joinRule) {
                    case "add":
                        double val = 0;
                        for (String key : jo.keySet()) {
                            val += jo.get(key).getAsDouble();
                        }
                        finalResult.result = new JsonPrimitive(val);
                        break;
                    case "multiply":
                        val = 1;
                        for (String key : jo.keySet()) {
                            val *= jo.get(key).getAsDouble();
                        }
                        finalResult.result = new JsonPrimitive(val);
                        break;
                }
            }
        }
    }
}
